import abc
import enum
import struct
import typing
from struct import Struct
from collections import deque
from .exceptions import ParseError
from . import read_raw_struct, read, read_until, read_struct, Parser, get_parser, wait

_parent_stack = deque()
_mapping_stack = deque()


class Unit(abc.ABC):
    @abc.abstractmethod
    def get_value(self) -> typing.Generator:
        "get object you want from bytes"

    @abc.abstractmethod
    def __call__(self, obj: typing.Any) -> bytes:
        "convert user-given object to bytes"

    def parse(self, data: bytes, *, strict=True):
        return Parser(self.get_value()).parse(data, strict=strict)


class BinarySchemaMetaclass(type):
    # def __new__(mcls, name, bases, namespace, **kwargs):
    #     fields = {}
    #     for key, value in namespace.items():
    #         if isinstance(value, (Unit, mcls)):
    #             fields[key] = value
    #     namespace["_fields"] = fields
    #     return super().__new__(mcls, name, bases, namespace)

    def __init__(cls, name, bases, namespace):
        fields = {}
        for key, value in namespace.items():
            if isinstance(value, (Unit, BinarySchemaMetaclass)):
                fields[key] = value
        cls._fields = fields
        super().__init__(name, bases, namespace)

    def __str__(cls):
        sl = []
        for name, field in cls._fields.items():
            sl.append(f"{name}={field}")
        s = ", ".join(sl)
        return f"{cls.__name__}({s})"


class BinarySchema(metaclass=BinarySchemaMetaclass):
    def __init__(self, *args):
        if len(args) != len(self.__class__._fields):
            raise ValueError(
                f"need {len(self.__class__._fields)} args, got {len(args)}"
            )
        _parent_stack.append(self)

        self.bins = {}
        for arg, (name, field) in zip(args, self.__class__._fields.items()):
            if isinstance(field, BinarySchemaMetaclass):
                binary = arg.binary
            elif isinstance(field, Unit):
                binary = field(arg)
            if arg is ...:
                arg = field.get_default()
            setattr(self, name, arg)
            self.bins[name] = binary
        self.binary = b"".join(self.bins.values())

        _parent_stack.pop()

    def __str__(self):
        sl = []
        for name in self.__class__._fields:
            value = getattr(self, name)
            sl.append(f"{name}={value!r}")
        s = ", ".join(sl)
        return f"{self.__class__.__name__}({s})"

    def __repr__(self):
        return f"<{self}>"

    def __eq__(self, other: "BinarySchema") -> bool:
        if not isinstance(other, self.__class__):
            return False
        for name in self.__class__._fields:
            if getattr(self, name) != getattr(other, name):
                return False
        return True

    @classmethod
    def get_value(cls) -> typing.Generator:
        mapping = {}
        _mapping_stack.append(mapping)
        try:
            for name, field in cls._fields.items():
                mapping[name] = yield from field.get_value()
        except Exception:
            raise ParseError(mapping)
        _mapping_stack.pop()
        return cls(*mapping.values())

    @classmethod
    def get_parser(cls) -> Parser:
        return Parser(cls.get_value())

    @classmethod
    def parse(cls, data: bytes, *, strict=True) -> "BinarySchema":
        return cls.get_parser().parse(data, strict=strict)


FieldType = typing.Union[BinarySchemaMetaclass, Unit]


class StructUnit(Unit):
    def __init__(self, format_: str):
        self._struct = Struct(format_)

    def __str__(self):
        return f"{self.__class__.__name__}({self._struct.format})"

    def get_value(self):
        return (yield from read_raw_struct(self._struct))[0]

    def __call__(self, obj) -> bytes:
        return self._struct.pack(obj)


class IntUnit(Unit):
    def __init__(self, length: int, byteorder: str, signed: bool = False):
        self.length = length
        self.byteorder = byteorder
        self.signed = signed

    def get_value(self):
        a = yield from read(self.length)
        return int.from_bytes(a, self.byteorder, signed=self.signed)

    def __call__(self, obj: int) -> bytes:
        return obj.to_bytes(self.length, self.byteorder, signed=self.signed)


int8 = StructUnit("b")
uint8 = StructUnit("B")
int16 = StructUnit("h")
int16be = StructUnit(">h")
uint16 = StructUnit("H")
uint16be = StructUnit(">H")
int24 = IntUnit(3, "little", signed=True)
int24be = IntUnit(3, "big", signed=True)
uint24 = IntUnit(3, "little", signed=False)
uint24be = IntUnit(3, "big", signed=False)
int32 = StructUnit("i")
int32be = StructUnit(">i")
uint32 = StructUnit("I")
uint32be = StructUnit(">I")
int64 = StructUnit("q")
int64be = StructUnit(">q")
uint64 = StructUnit("Q")
uint64be = StructUnit(">Q")
float32 = StructUnit("f")
float32be = StructUnit(">f")
float64 = StructUnit("d")
float64be = StructUnit(">d")


class Bytes(Unit):
    def __init__(self, length):
        self.length = length
        if length >= 0:
            self._struct = Struct(f"{length}s")

    def __str__(self):
        return f"{self.__class__.__name__}({self.length})"

    def get_value(self):
        if self.length >= 0:
            return (yield from read_raw_struct(self._struct))[0]
        else:
            return (yield from read())

    def __call__(self, obj) -> bytes:
        if self.length >= 0:
            return self._struct.pack(obj)
        else:
            return obj


class String(Bytes):
    def __init__(self, length, encoding="utf-8"):
        super().__init__(length)
        self.encoding = encoding

    def __str__(self):
        return f"{self.__class__.__name__}({self.length})"

    def get_value(self):
        v, = yield from read_raw_struct(self._struct)
        return v.decode(self.encoding)

    def __call__(self, obj: str) -> bytes:
        return super().__call__(obj.encode(self.encoding))


class MustEqual(Unit):
    def __init__(self, unit, value):
        self.unit = unit
        self.value = value

    def __str__(self):
        return f"{self.__class__.__name__}({self.unit}, {self.value})"

    def get_value(self):
        result = yield from self.unit.get_value()
        if self.value != result:
            raise ValueError(f"expect {self.value}, got {result}")
        return result

    def get_default(self):
        return self.value

    def __call__(self, obj) -> bytes:
        if obj is not ...:
            if self.value != obj:
                raise ValueError(f"expect {self.value}, got {obj}")
        return self.unit(self.value)


class EndWith(Unit):
    def __init__(self, _bytes):
        self._bytes = _bytes

    def __str__(self):
        return f"{self.__class__.__name__}({self._bytes})"

    def get_value(self):
        return (yield from read_until(self._bytes, return_tail=False))

    def __call__(self, obj: bytes) -> bytes:
        return obj + self._bytes


class LengthPrefixedBytes(Unit):
    def __init__(self, length_unit: StructUnit):
        self.length_unit = length_unit

    def __str__(self):
        return f"{self.__class__.__name__}({self.length_unit})"

    def get_value(self):
        length = yield from self.length_unit.get_value()
        return (yield from read_struct(f"{length}s"))[0]

    def __call__(self, obj: bytes) -> bytes:
        length = len(obj)
        return self.length_unit(length) + struct.pack(f"{length}s", obj)


class LengthPrefixedString(Unit):
    def __init__(self, length_unit: StructUnit, encoding="utf-8"):
        self.length_unit = length_unit
        self.encoding = encoding

    def __str__(self):
        return f"{self.__class__.__name__}({self.length_unit}, {self.encoding})"

    def get_value(self):
        length = yield from self.length_unit.get_value()
        v, = yield from read_struct(f"{length}s")
        return v.decode(self.encoding)

    def __call__(self, obj: str) -> bytes:
        length = len(obj)
        return self.length_unit(length) + struct.pack(
            f"{length}s", obj.encode(self.encoding)
        )


class LengthPrefixedObjectList(Unit):
    def __init__(self, length_unit: StructUnit, object_unit: FieldType):
        self.length_unit = length_unit
        self.object_unit = object_unit

    def __str__(self):
        return f"{self.__class__.__name__}({self.length_unit}, {self.object_unit})"

    def get_value(self):
        length = yield from self.length_unit.get_value()
        data, = yield from read_struct(f"{length}s")
        parser = Parser(self._gen())
        return parser.parse(data)

    def _gen(self):
        parser = yield from get_parser()
        lst = []
        yield from wait()
        while parser.has_more_data():
            lst.append((yield from self.object_unit.get_value()))
        return lst

    def __call__(self, obj_list: typing.List[FieldType]) -> bytes:
        if isinstance(self.object_unit, BinarySchemaMetaclass):
            bytes_ = b"".join(bs.binary for bs in obj_list)
        elif isinstance(self.object_unit, Unit):
            bytes_ = b"".join(self.object_unit(bs) for bs in obj_list)
        return self.length_unit(len(bytes_)) + bytes_


class LengthPrefixedObject(LengthPrefixedObjectList):
    def _gen(self):
        parser = yield from get_parser()
        v = yield from self.object_unit.get_value()
        if parser.has_more_data():
            raise ValueError("extra bytes left")
        return v

    def __call__(self, obj: FieldType) -> bytes:
        bytes_ = (
            obj.binary
            if isinstance(self.object_unit, BinarySchemaMetaclass)
            else self.object_unit(obj)
        )
        return self.length_unit(len(bytes_)) + bytes_


class Switch(Unit):
    def __init__(self, ref: str, cases: typing.Mapping[typing.Any, FieldType]):
        self.ref = ref
        self.cases = cases

    def __str__(self):
        return f"{self.__class__.__name__}({self.ref}, {self.cases})"

    def get_value(self):
        mapping = _mapping_stack[-1]
        unit = self.cases[mapping[self.ref]]
        return (yield from unit.get_value())

    def __call__(self, obj) -> bytes:
        parent = _parent_stack[-1]
        real_field = self.cases[getattr(parent, self.ref)]
        return real_field(obj) if isinstance(real_field, Unit) else obj.binary


class SizedIntEnum(Unit):
    def __init__(self, size_unit: StructUnit, enum_class):
        self.size_unit = size_unit
        self.enum_class = enum_class

    def __str__(self):
        return f"{self.__class__.__name__}({self.size_unit}, {self.enum_class})"

    def get_value(self):
        v = yield from self.size_unit.get_value()
        return self.enum_class(v)

    def __call__(self, obj: enum.IntEnum) -> bytes:
        return self.size_unit(obj.value)


class Convert(Unit):
    def __init__(
        self, unit: FieldType, *, encode: typing.Callable, decode: typing.Callable
    ):
        self.unit = unit
        self.encode = encode
        self.decode = decode

    def get_value(self):
        v = yield from self.unit.get_value()
        return self.decode(v)

    def __call__(self, obj) -> bytes:
        return self.unit(self.encode(obj))


def Group(**fields: typing.Dict[str, FieldType]) -> BinarySchema:
    return type("Group", (BinarySchema,), fields)



"io-free stream parser which helps implementing network protocols the `Sans-IO` way"
import typing
import warnings
from struct import Struct
from enum import IntEnum, auto
from collections import deque
from .exceptions import NoResult, ParseError

__version__ = "0.2.1"
_wait = object()


class Traps(IntEnum):
    _read = auto()
    _read_more = auto()
    _read_until = auto()
    _read_struct = auto()
    _read_int = auto()
    _write = auto()
    _wait = auto()
    _peek = auto()
    _get_parser = auto()


class State(IntEnum):
    _state_wait = auto()
    _state_next = auto()
    _state_end = auto()


class Parser:
    def __init__(self, gen: typing.Generator):
        self.gen = gen
        self.input = bytearray()
        self.output = bytearray()
        self.res_queue = deque()
        self._next_value = None
        self._last_trap = None
        self._pos = 0
        self._state = State._state_wait
        self._process()

    def parse(self, data: bytes, *, strict: bool = True) -> typing.Any:
        """
        parse bytes
        """
        self.send(data)
        if strict and self.has_more_data():
            raise ParseError("redundant data left")
        return self.get_result()

    def send(self, data: bytes = b"") -> None:
        """
        send data for parsing
        """
        self.input.extend(data)
        self._process()

    def read_output(self, nbytes: int = 0) -> bytes:
        """
        read *at most* ``nbytes``
        """
        if nbytes == 0 or len(self.output) <= nbytes:
            data = bytes(self.output)
            del self.output[:]
            return data
        data = bytes(self.output[:nbytes])
        del self.output[:nbytes]
        return data

    def read(self, nbytes: int = 0) -> bytes:
        "backward-compatible for v0.1.x"
        warnings.warn(
            "backward-compatible for v0.1.x, use read_output instead",
            DeprecationWarning,
        )
        return self.read_output(nbytes)

    @property
    def has_result(self) -> bool:
        return len(self.res_queue) > 0

    def get_result(self):
        """
        raises *NoResult* exception if no result has been set
        """
        self._process()
        if not self.has_result:
            raise NoResult("no result")
        return self.res_queue.popleft()

    def set_result(self, result):
        self.res_queue.append(result)

    def finished(self) -> bool:
        return self._state is State._state_end

    def _process(self):
        if self._state is State._state_end:
            return
        self._state = State._state_next
        while self._state is State._state_next:
            self._next_state()

    def _next_state(self) -> None:
        if self._last_trap is None:
            try:
                trap, *args = self.gen.send(self._next_value)
            except StopIteration as e:
                self._state = State._state_end
                self.set_result(e.value)
                return
            except Exception:
                self._state = State._state_end
                raise ParseError(
                    f"on value {self._next_value!r}"
                )
            else:
                if not isinstance(trap, Traps):
                    self._state = State._state_end
                    raise RuntimeError(f"Expect Traps object, but got: {trap}")
        else:
            trap, *args = self._last_trap
        result = getattr(self, trap.name)(*args)
        if result is _wait:
            self._state = State._state_wait
            self._last_trap = (trap, *args)
        else:
            self._state = State._state_next
            self._next_value = result
            self._last_trap = None

    def readall(self) -> bytes:
        """
        retrieve data from input back
        """
        return self._read(0)

    def has_more_data(self) -> bool:
        return len(self.input) > 0

    def write(self, data: bytes) -> None:
        self.output.extend(data)

    def _write(self, data: bytes) -> None:
        self.output.extend(data)

    def _wait(self) -> None:
        if not getattr(self, "_waiting", False):
            self._waiting = True
            return _wait
        self._waiting = False
        return

    def _read(self, nbytes: int = 0, from_=None) -> bytes:
        buf = self.input if from_ is None else from_
        if nbytes == 0:
            data = bytes(buf)
            del buf[:]
            return data
        if len(buf) < nbytes:
            return _wait
        data = bytes(buf[:nbytes])
        del buf[:nbytes]
        return data

    def _read_more(self, nbytes: int = 1, from_=None) -> bytes:
        buf = self.input if from_ is None else from_
        if len(buf) < nbytes:
            return _wait
        data = bytes(buf)
        del buf[:]
        return data

    def _read_until(self, data: bytes, return_tail: bool = True, from_=None) -> bytes:
        buf = self.input if from_ is None else from_
        index = buf.find(data, self._pos)
        if index == -1:
            self._pos = len(buf) - len(data) + 1
            self._pos = self._pos if self._pos > 0 else 0
            return _wait
        size = index + len(data)
        if return_tail:
            data = bytes(buf[:size])
        else:
            data = bytes(buf[:index])
        del buf[:size]
        self._pos = 0
        return data

    def _read_struct(self, struct_obj: Struct, from_=None) -> tuple:
        buf = self.input if from_ is None else from_
        size = struct_obj.size
        if len(buf) < size:
            return _wait
        result = struct_obj.unpack_from(buf)
        del buf[:size]
        return result

    def _read_int(self, nbytes: int, byteorder: str = "big", from_=None) -> int:
        buf = self.input if from_ is None else from_
        if len(buf) < nbytes:
            return _wait
        data = self._read(nbytes)
        return int.from_bytes(data, byteorder)

    def _peek(self, nbytes: int = 1, from_=None) -> bytes:
        buf = self.input if from_ is None else from_
        if len(buf) < nbytes:
            return _wait
        return bytes(buf[:nbytes])

    def _get_parser(self) -> "Parser":
        return self


def read(nbytes: int = 0, *, from_=None) -> bytes:
    """
    if nbytes = 0, read as many as it can, empty bytes is valid;
    if nbytes > 0, read *exactly* ``nbytes``
    """
    return (yield (Traps._read, nbytes, from_))


def read_more(nbytes: int = 1, *, from_=None) -> bytes:
    """
    read *at least* ``nbytes``
    """
    if nbytes <= 0:
        raise ValueError(f"nbytes must > 0, but got {nbytes}")
    return (yield (Traps._read_more, nbytes, from_))


def read_until(data: bytes, *, return_tail: bool = True, from_=None) -> bytes:
    """
    read until some bytes appear
    """
    return (yield (Traps._read_until, data, return_tail, from_))


def read_struct(fmt: str, *, from_=None) -> tuple:
    """
    read specific formatted data
    """
    return (yield (Traps._read_struct, Struct(fmt), from_))


def read_raw_struct(struct_obj: Struct, *, from_=None) -> tuple:
    """
    read raw struct formatted data
    """
    return (yield (Traps._read_struct, struct_obj, from_))


def read_int(nbytes: int, *, byteorder: str = "big", from_=None) -> int:
    """
    read some bytes as integer
    """
    if nbytes <= 0:
        raise ValueError(f"nbytes must > 0, but got {nbytes}")
    return (yield (Traps._read_int, nbytes, byteorder, from_))


def write(data: bytes) -> None:
    """
    write some bytes to the output buffer
    """
    return (yield (Traps._write, data))


def wait() -> None:
    """
    wait for next send or get_result event
    """
    return (yield (Traps._wait,))


def peek(nbytes: int = 1, *, from_=None) -> bytes:
    """
    peek many bytes without taking them away from buffer
    """
    if nbytes <= 0:
        raise ValueError(f"nbytes must > 0, but got {nbytes}")
    return (yield (Traps._peek, nbytes, from_))


def get_parser() -> Parser:
    return (yield (Traps._get_parser,))


def parser(generator_func: typing.Generator) -> typing.Generator:
    def create_parser(*args, **kwargs) -> Parser:
        return Parser(generator_func(*args, **kwargs))

    generator_func.parser = create_parser
    return generator_func
